#
#  Created by Stefan Richter on 2013-01-12.
#  Copyright (c) 2013 figo GmbH. All rights reserved.
#

import base64
from datetime import datetime, timedelta
import hashlib
import json
import logging
import re
import socket
import ssl
import sys
import urllib

if sys.version_info[0] > 2:
    import http.client as httplib
    from itertools import zip_longest as izip_longest
else:
    import httplib
    from itertools import izip_longest


from .models import *


logger = logging.getLogger(__name__)


class VerifiedHTTPSConnection(httplib.HTTPSConnection):

    """HTTPSConnection supporting certificate authentication based on fingerprint"""

    VALID_FINGERPRINTS = ("3A:62:54:4D:86:B4:34:38:EA:34:64:4E:95:10:A9:FF:37:27:69:C0",)

    def connect(self):
        # overrides the version in httplib so that we do certificate verification
        if sys.hexversion >= 0x02070000:
            sock = socket.create_connection((self.host, self.port), self.timeout, self.source_address)
        else:
            sock = socket.create_connection((self.host, self.port), self.timeout)
        if self._tunnel_host:
            self.sock = sock
            self._tunnel()

        # wrap the socket
        self.sock = ssl.wrap_socket(sock, self.key_file, self.cert_file)

        # verify the certificate fingerprint
        certificate = self.sock.getpeercert(True)
        if certificate is None:
            raise ssl.SSLError("Certificate validation failed")
        else:
            fingerprint = hashlib.sha1(certificate).hexdigest()
            fingerprint = ":".join(["".join(x) for x in izip_longest(*[iter(fingerprint.upper())] * 2)])
            if not fingerprint in VerifiedHTTPSConnection.VALID_FINGERPRINTS:
                raise ssl.SSLError("Certificate validation failed")


class FigoObject(object):

    API_ENDPOINT = "api.figo.me"
    API_SECURE = True

    def _query_api(self, path, data=None, method="GET"):
        """Helper method for making a REST-compliant API call

        :Parameters:
         - `path` - path on the server to call
         - `data` - Dictionary of data to send to the server in message body
         - `method` - HTTP verb to use for the request

        :Returns:
            the JSON-parsed result body
        """

        connection = VerifiedHTTPSConnection(
            self.API_ENDPOINT) if self.API_SECURE else httplib.HTTPConnection(self.API_ENDPOINT)
        connection.request(method, path, None if data is None else json.dumps(data),
                           {'Authorization': "Bearer %s" % self.access_token, 'Accept': 'application/json', 'Content-Type': 'application/json'})
        response = connection.getresponse()

        if response.status >= 200 and response.status < 300:
            response_data = response.read().decode("utf-8")
            if response_data == "":
                return {}
            return json.loads(response_data)
        elif response.status == 400:
            response_data = response.read().decode("utf-8")
            return json.loads(response_data)
        elif response.status == 401:
            return {'error': "unauthorized", 'error_description': "Missing, invalid or expired access token."}
        elif response.status == 403:
            return {'error': "forbidden", 'error_description': "Insufficient permission."}
        elif response.status == 404:
            return None
        elif response.status == 405:
            return {'error': "method_not_allowed", 'error_description': "Unexpected request method."}
        elif response.status == 503:
            return {'error': "service_unavailable", 'error_description': "Exceeded rate limit."}
        else:
            logger.warn("Querying the API failed when accessing '%s': %d", path, response.status)
            return {'error': "internal_server_error", 'error_description': "We are very sorry, but something went wrong"}

    def _query_api_with_exception(self, path, data=None, method="GET"):
        """Helper method analog to _query_api but raises an exception instead of simply returning"""
        response = self._query_api(path, data, method)
        if 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return response


class FigoException(Exception):

    """Base class for all exceptions transported via the figo connect API.

    They consist of a code-like `error` and a human readable `error_description`.
    """

    def __init__(self, error, error_description):
        super(FigoException, self).__init__()

        self.error = error
        self.error_description = error_description

    def __str__(self):
        return "FigoException: %s(%s)" % (repr(self.error_description), repr(self.error))

    @classmethod
    def from_dict(cls, dictionary):
        """Helper function creating an exception instance from the dictionary returned by the server"""
        return cls(dictionary['error'], dictionary['error_description'])


class FigoConnection(FigoObject):

    """Representing a not user-bound connection to the figo connect API.

    Its main purpose is to let user login via the OAuth2 API.
    """

    def __init__(self, client_id, client_secret, redirect_uri):
        """Creates a FigoConnection instance.

        :Parameters:
         - `client_id` - the OAuth Client ID as provided by your figo developer contact
         - `client_secret` - the OAuth Client Secret as provided by your figo developer contact
         - `redirect_uri` - the URI the users gets redirected to after the login is finished or if he presses cancels
        """

        self.client_id = client_id
        self.client_secret = client_secret
        self.redirect_uri = redirect_uri

    def _query_api(self, path, data=None):
        """Helper method for making a OAuth2-compliant API call

        :Parameters:
         - `path` - path on the server to call
         - `data` - Dictionary of data to send to the server in message body

        :Returns:
            the JSON-parsed result body
        """

        connection = VerifiedHTTPSConnection(self.API_ENDPOINT) if self.API_SECURE else httplib.HTTPConnection(self.API_ENDPOINT)
        connection.request("POST", path, urllib.urlencode(data),
                           {'Authorization': "Basic %s" % base64.b64encode(self.client_id + ":" + self.client_secret),
                            'Accept': 'application/json', 'Content-Type': 'application/x-www-form-urlencoded'})
        response = connection.getresponse()

        if response.status >= 200 and response.status < 300:
            response_data = response.read().decode("utf-8")
            if response_data == "":
                return {}
            return json.loads(response_data)
        elif response.status == 400:
            response_data = response.read().decode("utf-8")
            return json.loads(response_data)
        elif response.status == 401:
            return {'error': "unauthorized", 'error_description': "Missing, invalid or expired access token."}
        elif response.status == 403:
            return {'error': "forbidden", 'error_description': "Insufficient permission."}
        elif response.status == 404:
            return None
        elif response.status == 405:
            return {'error': "method_not_allowed", 'error_description': "Unexpected request method."}
        elif response.status == 503:
            return {'error': "service_unavailable", 'error_description': "Exceeded rate limit."}
        else:
            logger.warn("Querying the API failed when accessing '%s': %d", path, response.status)
            return {'error': "internal_server_error", 'error_description': "We are very sorry, but something went wrong"}

    def login_url(self, scope, state):
        """The URL a user should open in his/her web browser to start the login process.

        When the process is completed, the user is redirected to the URL provided to the constructor and passes on an authentication code. This code can be converted into an access token for data access.

        :Parameters:
         - `scope` - Scope of data access to ask the user for, e.g. `accounts=ro`
         - `state` - String passed on through the complete login process and to the redirect target at the end. It should be used to validated the authenticity of the call to the redirect URL

        :Returns:
            the URL of the first page of the login process
        """
        return ("https://" if self.API_SECURE else "http://") + self.API_ENDPOINT + "/auth/code?" + urllib.urlencode({'response_type': 'code', 'client_id': self.client_id, 'redirect_uri': self.redirect_uri, 'scope': scope, 'state': state})

    def convert_authentication_code(self, authentication_code):
        """Convert the authentication code received as result of the login process into an access token usable for data access.

        :parameters:
         - `authentication_code` - the code received as part of the call to the redirect URL at the end of the logon process

        :returns:
            Dictionary with the following keys:
             - `access_token` - the access token for data access. You can pass it into `FigoConnection.open_session` to get a FigoSession and access the users data
             - `refresh_token` - if the scope contained the `offline` flag, also a refresh token is generated. It can be used to generate new access tokens, when the first one has expired.
             - `expires` - absolute time the access token expires
        """

        if authentication_code[0] != "O":
            raise Exception("Invalid authentication code")

        response = self._query_api("/auth/token", data={
                                   'code': authentication_code, 'redirect_uri': self.redirect_uri, 'grant_type': 'authorization_code'})
        if 'error' in response:
            raise FigoException.from_dict(response)

        return {'access_token': response['access_token'],
                'refresh_token': response['refresh_token'] if 'refresh_token' in response else None,
                'expires': datetime.now() + timedelta(seconds=response['expires_in'])}

    def convert_refresh_token(self, refresh_token):
        """Convert a refresh token (granted for offline access and returned by `convert_authentication_code`) into an access token usabel for data acccess.

        :Parameters:
         - `refresh_token` - refresh token returned by `convert_authentication_code`

        :Returns:
            Dictionary with the following keys:
             - `access_token` - the access token for data access. You can pass it into `FigoConnection.open_session` to get a FigoSession and access the users data
             - `expires` - absolute time the access token expires
        """

        if refresh_token[0] != "R":
            raise Exception("Invalid refresh token")

        response = self._query_api("/auth/token", data={
                                   'refresh_token': refresh_token, 'redirect_uri': self.redirect_uri, 'grant_type': 'refresh_token'})
        if 'error' in response:
            raise FigoException.from_dict(response)

        return {'access_token': response['access_token'],
                'expires': datetime.now() + timedelta(seconds=response['expires_in'])}

    def revoke_token(self, token):
        """Revoke a granted access or refresh token and thereby invalidate it.

        Note: this action has immediate effect, i.e. you will not be able use that token anymore after this call.

        :Parameters:
         - `token` - access or refresh token to be revoked
        """

        response = self._query_api("/auth/revoke?" + urllib.urlencode({'token': token}))
        if 'error' in response:
            raise FigoException.from_dict(response)

    def add_user(self, **kwargs):
        """Register a new figo Account. Only native clients may create new figo Accounts.

        :Parameters:
            all user attributes as keyword arguments

        :Returns:
            Auto-generated recovery password.
        """

        response = self._query_api("/auth/user", kwargs, method="POST")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return response['recovery_password']

    def resend_unlock_code(self, **kwargs):
        """Re-send unlock code email.

        :Parameters:
            all resend unlock'code attributes as keyword arguments

        """

        response = self._query_api("/auth/user/resend_unlock_code", kwargs, method="POST")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def unlock(self, **kwargs):
        """Native clients may unlock figo Accounts. This request must be issued with the same client identifier which was used to request the unlock code.

        :Parameters:
            all unlock attributes as keyword arguments

        :Returns:
            Auto-generated recovery password.
        """

        response = self._query_api("/auth/unlock", kwargs, method="POST")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)


class FigoSession(FigoObject):

    """Represents a user-bound connection to the figo connect API and allows access to the users data"""

    def __init__(self, access_token):
        """Creates a FigoSession instance.

        :Parameters:
         - `access_token` - the access token to bind this session to a user
        """
        self.access_token = access_token

    @property
    def accounts(self):
        """An array of `Account` objects, one for each account the user has granted the app access"""

        response = self._query_api("/rest/accounts")
        if 'error' in response:
            raise FigoException.from_dict(response)
        return [Account.from_dict(self, account_dict) for account_dict in response['accounts']]

    def get_account(self, account_id):
        """Retrieve a specific account.

        :Parameters:
         - `account_id` - ID of the account to be retrieved

        :Returns:
            `Account` object for the respective account
        """

        response = self._query_api("/rest/accounts/%s" % account_id)
        if 'error' in response:
            raise FigoException.from_dict(response)
        return Account.from_dict(self, response)

    def add_account(self, **kwargs):
        """Create a new account.

        :Parameters:
            all account attributes as keyword arguments

        :Returns:
            Task token
        """

        response = self._query_api("/rest/accounts", kwargs, method="POST")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return response['task_token']

    def modify_account_sort_order(self, **kwargs):
        """Set account sort order.

        :Parameters:
         - all accounts attributes as keyword arguments
        """

        response = self._query_api("/rest/accounts", kwargs, method="PUT")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def modify_account(self, account_id, **kwargs):
        """Modify an account.

        :Parameters:
         - `account_id` - ID of the account to be modified
         - all account attributes as keyword arguments
        """

        response = self._query_api("/rest/accounts/%s" % account_id, kwargs, method="PUT")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def remove_account(self, account_id):
        """Remove an account

        :Parameters:
         - `account_id` - ID of the account to be deleted
        """

        response = self._query_api("/rest/accounts/%s" % account_id, method="DELETE")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def get_account_balance(self, account_id):
        """Get balance and account limits.

        :Parameters:
         - `account_id` - ID of the account to be retrieved

        :Returns:
            `AccountBalance` object for the respective account
        """

        response = self._query_api("/rest/accounts/%s/balance" % account_id)
        if 'error' in response:
            raise FigoException.from_dict(response)
        return AccountBalance.from_dict(self, response)

    def modify_account_balance(self, account_id, **kwargs):
        """Modify balance or account limits.

        :Parameters:
         - `account_id` - ID of the account to be modified
         - all account balance attributes as keyword arguments
        """

        response = self._query_api("/rest/accounts/%s/balance" % account_id, kwargs, method="PUT")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    @property
    def catalog_services(self):
        """An array of `Service` objects, one for each supported credit card and payment services."""

        response = self._query_api("/rest/catalog/services/de")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return [Service.from_dict(self, service_dict) for service_dict in response['services']]

    def get_catalog_bank(self, bank_code, country_code='de'):
        """Get login settings for bank code.
        :Parameters:
         - `bank_code` - Bank code  

        :Returns:
            'Bank' object for the respective bank
        """
        response = self._query_api("/rest/catalog/banks/%s/%s" % (country_code, bank_code))
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return Bank.from_dict(self, response)

    @property
    def clients(self):
        """Get a list of all current and former authorized figo Connect clients."""

        response = self._query_api("/rest/clients")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return [Client.from_dict(self, client_dict) for client_dict in response['clients']]

    def modify_client(self, client_id, **kwargs):
        """Modify client details.

        :Parameters:
         - `client_id` - Internal figo Connect client ID.
         - all Client attributes as keyword arguments
        
        """

        response = self._query_api("/rest/clients/%s" % client_id, kwargs, "PUT")

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def remove_client(self, client_id):
        """Unlink client.

        :Parameters:
         - `client_id` -Internal figo Connect client ID.
        """

        response = self._query_api("/rest/clients/%s" (client_id), method="DELETE")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    @property
    def devices(self):
        """Get a list of all authorized devices."""

        response = self._query_api("/rest/devices")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return [Device.from_dict(self, device_dict) for device_dict in response['devices']]

    def modify_device(self, device_id, **kwargs):
        """Modify client details.

        :Parameters:
         - `device_id` - Internal figo Connect device ID.
         - all Device attributes as keyword arguments
        
        """

        response = self._query_api("/rest/devices/%s" % device_id, kwargs, "PUT")

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def remove_device(self, device_id):
        """Unlink client.

        :Parameters:
         - `device_id` - Internal figo Connect device ID.
        """

        response = self._query_api("/rest/devices/%s" (device_id), method="DELETE")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    @property
    def notifications(self):
        """An array of `Notification` objects, one for each registered notification"""

        response = self._query_api("/rest/notifications")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return [Notification.from_dict(self, notification_dict) for notification_dict in response['notifications']]

    def get_notification(self, notification_id):
        """Retrieve a specific notification.

        :Parameters:
         - `notification_id` - ID of the notification to be retrieved

        :Returns:
            'Notification' object for the respective notification
        """

        response = self._query_api("/rest/notifications/" + str(notification_id))
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return Notification.from_dict(self, response)

    def add_notification(self, **kwargs):
        """Create a new notification.

        :Parameters:
            all Notification attributes as keyword arguments

        :Returns:
            ID of the newly created notification
        """

        response = self._query_api("/rest/notifications", kwargs, method="POST")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return response['notification_id']

    def modify_notification(self, notification_id, **kwargs):
        """Modify a notification.

        :Parameters:
         - `notification_id` - ID of the notification to be modified
         - all Notification attributes as keyword arguments
        """

        response = self._query_api("/rest/notifications/" + str(notification_id), kwargs, method="PUT")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def remove_notification(self, notification_id):
        """Remove a notification

        :Parameters:
         - `notification_id` - ID of the notification to be deleted
        """

        response = self._query_api("/rest/notifications/" + str(notification_id), method="DELETE")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def get_payments(self, account_id, since=None, count=None):
        """Get an array of `Payment` objects, one for each payment of the user

        :Parameters:
         - `account_id` - ID of the account to be retrieved
         - `since` - Payments which were created on or after this date/id will be returned in the response
         - `count` - Limit the number of returned payments. The default number is 1000.

        :Returns:
            `List` of Payment objects
        """

        query = {}
        if since:
            query["since"] = since

        if count:
            query["count"] = count

        response = self._query_api("/rest/accounts/%s/payments" % (account_id), query)

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return [Payment.from_dict(self, payment_dict) for payment_dict in response['payments']]

    def get_scheduled_payments(self, account_id):
        """Get an array of `Payment` objects, one for each scheduled payments known to the bank server.

        :Parameters:
         - `account_id` - ID of the account to be retrieved

        :Returns:
            `List` of Payment objects
        """

        response = self._query_api("/rest/accounts/%s/payments/scheduled" % (account_id))

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return [Payment.from_dict(self, payment_dict) for payment_dict in response['payments']]

    def get_payment(self, account_id, payment_id):
        """Get a single `Payment` object

        :Parameters:
         - `account_id` - Internal figo Connect account ID
         - `payment_id` - ID of the payment to be retrieved
         
        :Returns:
            `Payment` object
        """

        response = self._query_api("/rest/accounts/%s/payments/%s" % (account_id, payment_id))

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return Payment.from_dict(self, response)

    def get_scheduled_payment(self, account_id, payment_id):
        """Get information about a particular scheduled payment. 

        :Parameters:
         - `account_id` - Internal figo Connect account ID
         - `payment_id` - ID of the payment to be retrieved
         
        :Returns:
            `Payment` object
        """

        response = self._query_api("/rest/accounts/%s/scheduled/%s" % (account_id, payment_id))

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return Payment.from_dict(self, response)

    def add_payment(self, account_id, **kwargs):
        """Insert payment into payment list

        :Parameters:
         - `account_id` - Internal figo Connect account ID
         - all Payment attributes as keyword arguments
        

        :Returns:
            `Payment` object of the new created Payment 
        """

        response = self._query_api("/rest/accounts/%s/payments" % (account_id), kwargs, "POST")

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return Payment.from_dict(self, response)

    def modify_payment(self, account_id, payment_id, **kwargs):
        """Modify a payment

        :Parameters:
         - `account_id` - Internal figo Connect account ID
         - `payment_id` - ID of the payment to be modified
         - all Payment attributes as keyword arguments
        

        """

        response = self._query_api("/rest/accounts/%s/payments/%s" % (account_id, payment_id), kwargs, "PUT")

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def modify_scheduled_payment(self, account_id, payment_id, **kwargs):
        """Modify a payment scheduled

        :Parameters:
         - `account_id` - Internal figo Connect account ID
         - `payment_id` - ID of the payment to be modified
         - all Payment attributes as keyword arguments
        

        :Returns:
            `task_token` Task token
        """

        response = self._query_api("/rest/accounts/%s/scheduled/%s" % (account_id, payment_id), kwargs, "PUT")

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return response["task_token"]

    def remove_payment(self, account_id, payment_id):
        """Remove a payment

        :Parameters:
         - `account_id` - Internal figo Connect account ID
         - `payment_id` - ID of the payment to be deleted
        """

        response = self._query_api("/rest/accounts/%s/payments/%s" (account_id, notification_id), method="DELETE")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def remove_scheduled_payment(self, account_id, payment_id, **kwargs):
        """Remove a scheduled payment

        :Parameters:
         - `account_id` - Internal figo Connect account ID
         - `payment_id` - ID of the payment to be deleted

         :Returns:
            `task_token` Task token
        """

        response = self._query_api("/rest/accounts/%s/scheduled/%s" (account_id, notification_id), kwargs, method="DELETE")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return response["task_token"]

    def submit_payment(self, account_id, payment_id, **kwargs):
        """Submit payment to bank server.

        :Parameters:
         - `account_id` - Internal figo Connect account ID
         - `payment_id` - ID of the payment to be submited
         - all Payment submit attributes as keyword arguments
        

        :Returns:
            `task_token` Task token
        """

        response = self._query_api("/rest/accounts/%s/payments/%s/submit" % (account_id, payment_id), kwargs, "POST")

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return response["task_token"]

    @property
    def transactions(self):
        """An array of `Transaction` objects, one for each transaction of the user"""

        response = self._query_api("/rest/transactions")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return [Transaction.from_dict(self, transaction_dict) for transaction_dict in response['transactions']]

    def get_transactions(self, account_id):
        """Get an array of `Transaction` objects, one for each transaction of the user

        :Parameters:
         - `account_id` - ID of the account to be retrieved
 
        :Returns:
            `List` of Transaction objects
        """

        response = self._query_api("/rest/accounts/%s/transactions" % (account_id))

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return [Transaction.from_dict(self, transaction_dict) for transaction_dict in response['transactions']]

    def get_transaction(self, account_id, transaction_id):
        """Retrieve a specific transaction.

        :Parameters:
         - `account_id` - ID of the account to be retrieved
         - `transaction_id` - ID of the transaction to be retrieved
 
        :Returns:
            a `Transaction` object representing the transaction to be retrieved
        """

        response = self._query_api("/rest/accounts/%s/transactions/%s" % (account_id, transaction_id))

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return Transaction.from_dict(self, response)

    def modify_transaction(self, account_id, transaction_id, **kwargs):
        """Modify a transaction

        :Parameters:
         - `account_id` - Internal figo Connect account ID
         - `transaction_id` - ID of the transaction to be modified
         - all Transaction attributes as keyword arguments
        
        """

        response = self._query_api("/rest/accounts/%s/transactions/%s" % (account_id, transaction_id), kwargs, "PUT")

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def remove_transaction(self, account_id, transaction_id):
        """Remove a Transaction

        :Parameters:
         - `account_id` - Internal figo Connect account ID
         - `transaction_id` - ID of the transaction to be deleted
        """

        response = self._query_api("/rest/accounts/%s/transactions/%s" (account_id, notification_id), method="DELETE")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def get_bank(self, bank_id):
        """Get bank contact settings.

        :Parameters:
         - `bank_id` - Internal figo Connect bank ID.
        

        :Returns:
            a `BankContact` object representing the bank contact to be retrieved
        """

        response = self._query_api("/rest/banks/%s" % (bank_id))

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return BankContact.from_dict(self, response)

    def modify_bank(self, bank_id, **kwargs):
        """Modify a transaction

        :Parameters:
         - `bank_id` - Internal figo Connect bank ID.
         - all BankContact attributes as keyword arguments
        
        """

        response = self._query_api("/rest/banks/%s" % (account_id, transaction_id), kwargs, "PUT")

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def remove_bank_pin(self, bank_id):
        """Remove a Transaction

        :Parameters:
        - `bank_id` - Internal figo Connect bank ID.
        """

        response = self._query_api("/rest/banks/%s/remove_pin" (bank_id), method="DELETE")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    @property
    def user(self):
        """Get figo Account settings."""

        response = self._query_api("/rest/user")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return User.from_dict(self, response)

    def modify_user(self, **kwargs):
        """Modify figo Account settings.

        :Parameters:
         - all User attributes as keyword arguments
        
        """

        response = self._query_api("/rest/user", kwargs, "PUT")

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def remove_user(self):
        """Delete figo Account.

        """

        response = self._query_api("/rest/user", method="DELETE")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def resend_verification(self):
        """Re-send verification email.

        """

        response = self._query_api("/rest/user/resend_verification", method="POST")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def purchase_premium(self, **kwargs):
        """Purchase a premium membership with a debit or credit card.


        :Parameters:
         - all purchase premium attributes as keyword arguments

         :Returns:
            a dict with the response

        """

        response = self._query_api("/rest/user/purchase_subscription", kwargs, method="POST")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return response

    def cancel_subscription(self):
        """Cancel the premium membership

        """

        response = self._query_api("/rest/user/cancel_subscription", method="POST")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def verify_receipt(self, **kwargs):
        """Verify iOS App Store receipt from In-App Purchase.


        :Parameters:
         - all verify receipt attributes as keyword arguments

         :Returns:
            a dict with the response

        """

        response = self._query_api("/rest/user/verify_receipt", kwargs, method="POST")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return response

    def redeem_code(self, **kwargs):
        """Redeem a voucher code.

        :Parameters:
         - all redeem code attributes as keyword arguments

         :Returns:
            a dict with the response

        """

        response = self._query_api("/rest/user/redeem_code", kwargs, method="POST")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return response

    def get_sync_url(self, state, redirect_uri):
        """URL to trigger a synchronisation.

        The user should open this URL in a web browser to synchronize his/her accounts with the respective bank servers. When the process is finished, the user is redirected to the provided URL.

        :Parameters:
         - `state` - String passed on through the complete synchronization process and to the redirect target at the end. It should be used to validated the authenticity of the call to the redirect URL
         - `redirect_uri` - URI the user is redirected to after the process completes

        :Returns:
            the URL to be opened by the user.
        """

        response = self._query_api("/rest/sync", {"state": state, "redirect_uri": redirect_uri}, method="POST")
        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return FigoConnection.API_ENDPOINT + "/task/start?id=" + response['task_token']

    def get_task_start_url(self, task_token):
        """URL for communication with bank server

        :Parameters:
         - `task_token` - Task token from the initial request.

        :Returns:
            the URL to be opened by the user.
        """
        return "%s/task/start?id=%s" % (FigoConnection.API_ENDPOINT, task_token)

    def task_progress(self, task_id, **kwargs):
        """While the figo Connect server communicates with a bank server, your application can monitor its progress by periodically polling this method.

        :Parameters:
         - `task_id` - Task token
         - all Task attributes as keyword arguments
        

        :Returns:
            `Task` object
        """

        response = self._query_api("/task/progress?id=%s" % task_id, kwargs, "POST")

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)
        else:
            return Task.from_dict(response)

    def cancel_task(self, task_id):
        """Cancel bank server communication.

        :Parameters:
         - `task_id` - Task token
         
        """
        response = self._query_api("/task/cancel?id=%s" % task_id, kwargs, "POST")

        if response is None:
            return None
        elif 'error' in response:
            raise FigoException.from_dict(response)

    def parse_webhook_notification(self, message_body):
        """ parses webhook notification and returns a `WebhookNotification` object

            :Parameters:
            - `message_body` - message body of the webhook message (as string or dict)

            :Returns:
                a WebhookNotification object 

        """

        if type(message_body) is not dict:
            message_body = json.loads(message_body)

        notification = WebhookNotification.from_dict(self, message_body)

        data = self._query_api(notification.observe_key)

        if re.match("\/rest\/transactions", notification.observe_key):
            notification.data = [Transaction.from_dict(self, transaction_dict) for transaction_dict in response['transactions']]

        elif re.match("\/rest\/accounts\/(.*)\/transactions", notification.observe_key):
            notification.data = [Transaction.from_dict(self, transaction_dict) for transaction_dict in response['transactions']]

        elif re.match("\/rest\/accounts\/(.*)\/balance", notification.observe_key):
            notification.data = AccountBalance.from_dict(data)

        return notification
